@@ -4,11 +4,24 @@ package TCOD;
 use strict;
 use warnings;
 
+use Encode ();
+use FFI::C;
 use FFI::CheckLib ();
 use FFI::Platypus 1.00;
-use FFI::C;
+use FFI::Platypus::Buffer ();
+use Ref::Util;
+use TCOD::SDL2;
+use TCOD::Event;
 
-our $VERSION = '0.002';
+sub import {
+    use Import::Into;
+    TCOD::Event->import::into( scalar caller );
+}
+
+our $VERSION = '0.003';
+
+my $bundle = FFI::Platypus->new( api => 1 );
+$bundle->bundle('TCOD');
 
 my $ffi;
 BEGIN {
@@ -19,6 +32,8 @@ BEGIN {
 
 $ffi->load_custom_type( '::WideString' => 'wstring', access => 'read' );
 
+$ffi->attach( [ TCOD_get_error => 'get_error' ] => [] => 'string' );
+
 sub enum {
     my %enums = @_;
     while ( my ( $name, $values ) = each %enums ) {
@@ -36,6 +51,75 @@ use constant {
     NOISE_MAX_DIMENSIONS     => 4,
     NOISE_DEFAULT_HURST      => 0.5,
     NOISE_DEFAULT_LACUNARITY => 2,
+
+    CHARMAP_CP437 => [
+        0x0000, 0x263A, 0x263B, 0x2665, 0x2666, 0x2663, 0x2660, 0x2022,
+        0x25D8, 0x25CB, 0x25D9, 0x2642, 0x2640, 0x266A, 0x266B, 0x263C,
+        0x25BA, 0x25C4, 0x2195, 0x203C, 0x00B6, 0x00A7, 0x25AC, 0x21A8,
+        0x2191, 0x2193, 0x2192, 0x2190, 0x221F, 0x2194, 0x25B2, 0x25BC,
+        0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
+        0x0028, 0x0029, 0x002A, 0x002B, 0x002C, 0x002D, 0x002E, 0x002F,
+        0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
+        0x0038, 0x0039, 0x003A, 0x003B, 0x003C, 0x003D, 0x003E, 0x003F,
+        0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
+        0x0048, 0x0049, 0x004A, 0x004B, 0x004C, 0x004D, 0x004E, 0x004F,
+        0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
+        0x0058, 0x0059, 0x005A, 0x005B, 0x005C, 0x005D, 0x005E, 0x005F,
+        0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
+        0x0068, 0x0069, 0x006A, 0x006B, 0x006C, 0x006D, 0x006E, 0x006F,
+        0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
+        0x0078, 0x0079, 0x007A, 0x007B, 0x007C, 0x007D, 0x007E, 0x007F,
+        0x00C7, 0x00FC, 0x00E9, 0x00E2, 0x00E4, 0x00E0, 0x00E5, 0x00E7,
+        0x00EA, 0x00EB, 0x00E8, 0x00EF, 0x00EE, 0x00EC, 0x00C4, 0x00C5,
+        0x00C9, 0x00E6, 0x00C6, 0x00F4, 0x00F6, 0x00F2, 0x00FB, 0x00F9,
+        0x00FF, 0x00D6, 0x00DC, 0x00A2, 0x00A3, 0x00A5, 0x20A7, 0x0192,
+        0x00E1, 0x00ED, 0x00F3, 0x00FA, 0x00F1, 0x00D1, 0x00AA, 0x00BA,
+        0x00BF, 0x2310, 0x00AC, 0x00BD, 0x00BC, 0x00A1, 0x00AB, 0x00BB,
+        0x2591, 0x2592, 0x2593, 0x2502, 0x2524, 0x2561, 0x2562, 0x2556,
+        0x2555, 0x2563, 0x2551, 0x2557, 0x255D, 0x255C, 0x255B, 0x2510,
+        0x2514, 0x2534, 0x252C, 0x251C, 0x2500, 0x253C, 0x255E, 0x255F,
+        0x255A, 0x2554, 0x2569, 0x2566, 0x2560, 0x2550, 0x256C, 0x2567,
+        0x2568, 0x2564, 0x2565, 0x2559, 0x2558, 0x2552, 0x2553, 0x256B,
+        0x256A, 0x2518, 0x250C, 0x2588, 0x2584, 0x258C, 0x2590, 0x2580,
+        0x03B1, 0x00DF, 0x0393, 0x03C0, 0x03A3, 0x03C3, 0x00B5, 0x03C4,
+        0x03A6, 0x0398, 0x03A9, 0x03B4, 0x221E, 0x03C6, 0x03B5, 0x2229,
+        0x2261, 0x00B1, 0x2265, 0x2264, 0x2320, 0x2321, 0x00F7, 0x2248,
+        0x00B0, 0x2219, 0x00B7, 0x221A, 0x207F, 0x00B2, 0x25A0, 0x00A0,
+    ],
+    CHARMAP_TCOD => [
+        0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
+        0x0028, 0x0029, 0x002A, 0x002B, 0x002C, 0x002D, 0x002E, 0x002F,
+        0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
+        0x0038, 0x0039, 0x003A, 0x003B, 0x003C, 0x003D, 0x003E, 0x003F,
+        0x0040, 0x005B, 0x005C, 0x005D, 0x005E, 0x005F, 0x0060, 0x007B,
+        0x007C, 0x007D, 0x007E, 0x2591, 0x2592, 0x2593, 0x2502, 0x2500,
+        0x253C, 0x2524, 0x2534, 0x251C, 0x252C, 0x2514, 0x250C, 0x2510,
+        0x2518, 0x2598, 0x259D, 0x2580, 0x2596, 0x259A, 0x2590, 0x2597,
+        0x2191, 0x2193, 0x2190, 0x2192, 0x25B2, 0x25BC, 0x25C4, 0x25BA,
+        0x2195, 0x2194, 0x2610, 0x2611, 0x25CB, 0x25C9, 0x2551, 0x2550,
+        0x256C, 0x2563, 0x2569, 0x2560, 0x2566, 0x255A, 0x2554, 0x2557,
+        0x255D, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
+        0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 0x0048,
+        0x0049, 0x004A, 0x004B, 0x004C, 0x004D, 0x004E, 0x004F, 0x0050,
+        0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 0x0058,
+        0x0059, 0x005A, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
+        0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 0x0068,
+        0x0069, 0x006A, 0x006B, 0x006C, 0x006D, 0x006E, 0x006F, 0x0070,
+        0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 0x0078,
+        0x0079, 0x007A, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
+        0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
+        0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
+        0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
+        0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
+        0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
+        0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
+        0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
+        0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
+        0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
+        0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
+        0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
+        0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
+    ],
 };
 
 BEGIN {
@@ -51,6 +135,14 @@ BEGIN {
         RIGHT                   =>   1,
         CENTER                  =>   2,
     },
+    Error => {
+        E_OK                    =>   0,
+        E_ERROR                 =>  -1,
+        E_INVALID_ARGUMENT      =>  -2,
+        E_OUT_OF_MEMORY         =>  -3,
+        E_REQUIRES_ATTENTION    =>  -4,
+        E_WARN                  =>   1,
+    },
     Renderer => {
         RENDERER_GLSL           =>   0,
         RENDERER_OPENGL         =>   1,
@@ -306,19 +398,23 @@ BEGIN {
 
 $ffi->type( int    => 'TCOD_renderer' );
 $ffi->type( int    => 'TCOD_keycode'  );
+$ffi->type( int    => 'TCOD_error'    );
 $ffi->type( opaque => 'TCOD_event'    );
 $ffi->type( '(int, int, int, int, opaque )->float' => 'TCOD_path_func' );
 
-# Blessed opaque types
-for my $name (qw( image console map path dijkstra random noise )) {
-    $ffi->custom_type( "TCOD_$name" => {
+# Custom blessed opaque types
+for my $name (qw( image console map path dijkstra random noise context )) {
+    my $type = {
         native_type    => 'opaque',
-        perl_to_native => sub { $_[0] ? ${ $_[0] } : undef    },
+        perl_to_native => sub { $_[0] ? ${ $_[0] } : undef },
         native_to_perl => sub {
             return unless $_[0];
             bless \$_[0], 'TCOD::' . ucfirst $name;
         },
-    });
+    };
+
+    $ffi->custom_type( "TCOD_$name" => $type );
+    $bundle->custom_type( "TCOD_$name" => $type );
 }
 
 package TCOD::Key {
@@ -408,7 +504,8 @@ package TCOD::Color {
 
     use FFI::Platypus::Record;
     record_layout_1( uint8 => 'r', uint8 => 'g', uint8 => 'b' );
-    $ffi->type( 'record(TCOD::Color)'  => 'TCOD_color'  );
+    $ffi->type(    'record(TCOD::Color)'  => 'TCOD_color'  );
+    $bundle->type( 'record(TCOD::Color)'  => 'TCOD_color'  );
 
     {
         # Give color a positional constructor
@@ -416,14 +513,15 @@ package TCOD::Color {
         no warnings 'redefine';
 
         my $old  = __PACKAGE__->can('new') or die;
-        my $new  = sub { shift->$old({ r => shift, g => shift, b => shift })  };
         my $name = __PACKAGE__ . '::new';
 
         require Sub::Util;
-        *{$name} = Sub::Util::set_subname $name => $new;
+        *{$name} = Sub::Util::set_subname $name => sub {
+            my $class = shift;
+            $class->$old({ r => shift, g => shift, b => shift });
+        };
     }
 
-
     $ffi->attach( equals          => [qw( TCOD_color TCOD_color            )] => 'bool'       );
     $ffi->attach( add             => [qw( TCOD_color TCOD_color            )] => 'TCOD_color' );
     $ffi->attach( subtract        => [qw( TCOD_color TCOD_color            )] => 'TCOD_color' );
@@ -469,6 +567,28 @@ package TCOD::Color {
     }
 }
 
+package TCOD::ColorRGBA {
+    use FFI::Platypus::Record;
+    record_layout_1( uint8 => 'r', uint8 => 'g', uint8 => 'b', uint8 => 'a' );
+    $ffi->type( 'record(TCOD::ColorRGBA)'  => 'TCOD_colorRGBA'  );
+    $bundle->type( 'record(TCOD::ColorRGBA)'  => 'TCOD_colorRGBA'  );
+
+    {
+        # Give color a positional constructor
+        no strict 'refs';
+        no warnings 'redefine';
+
+        my $old  = __PACKAGE__->can('new') or die;
+        my $name = __PACKAGE__ . '::new';
+
+        require Sub::Util;
+        *{$name} = Sub::Util::set_subname $name => sub {
+            my $class = shift;
+            $class->$old({ r => shift, g => shift, b => shift, a => shift });
+        };
+    }
+}
+
 package TCOD::Map {
     $ffi->mangler( sub { 'TCOD_map_' . shift } );
 
@@ -487,10 +607,10 @@ package TCOD::Map {
 }
 
 package TCOD::Console {
-    $ffi->mangler( sub { shift } );
-    $ffi->attach( [ TCOD_console_delete => 'DESTROY' ] => [qw( TCOD_console )] => 'void' );
+    $ffi->mangler(    sub { 'TCOD_console_' . shift } );
+    $bundle->mangler( sub { 'PERL_console_' . shift } );
 
-    $ffi->mangler( sub { 'TCOD_console_' . shift } );
+    $ffi->attach( [ delete => 'DESTROY' ] => [qw( TCOD_console )] => 'void' );
 
     # Constructors
     $ffi->attach( new       => [qw( int int )] => 'TCOD_console' => sub { $_[0]->( @_[ 2 .. $#_ ] ) } );
@@ -498,12 +618,165 @@ package TCOD::Console {
     $ffi->attach( load_asc  => [qw( string  )] => 'TCOD_console' => sub { $_[0]->( $_[ 2 ] ) } );
     $ffi->attach( load_apf  => [qw( string  )] => 'TCOD_console' => sub { $_[0]->( $_[ 2 ] ) } );
 
+    # Printing methods
+
+    my $blit_with_key_color = $ffi->function( blit_key_color => [qw(
+        TCOD_console int int int int
+        TCOD_console int int
+        float float
+        TCOD_color
+    )]);
+
+    $ffi->attach( blit => [qw(
+        TCOD_console int int int int
+        TCOD_console int int
+        float float
+    )] => void => sub {
+        my ( $xsub, $console, %args ) = @_;
+
+        my @args = (
+            $console,
+            $args{src_x}    // 0,
+            $args{src_y}    // 0,
+            $args{width}    // 0,
+            $args{height}   // 0,
+            $args{dest},
+            $args{dest_x}   // 0,
+            $args{dest_y}   // 0,
+            $args{fg_alpha} // $args{alpha} // 1,
+            $args{bg_alpha} // $args{alpha} // 1,
+        );
+
+        return $blit_with_key_color->( @args, $args{key_color} ) if $args{key_color};
+
+        $xsub->(@args);
+    });
+
+    $ffi->attach( [ printn => 'print' ] => [qw(
+        TCOD_console
+        int
+        int
+        size_t
+        string
+        TCOD_color*
+        TCOD_color*
+        int
+        int
+    )] => TCOD_error => sub {
+        my ( $xsub, $console, %args ) = ( shift, shift );
+
+        # Accommodate $con->print( $x, $y, $string, %rest );
+        %args = ( x => shift, y => shift, string => shift ) if @_ % 2;
+        %args = ( %args, @_ );
+
+        my $string = Encode::encode( 'UTF-8', $args{string}, Encode::FB_CROAK );
+
+        $xsub->(
+            $console,  @args{qw( x y )},
+            length($string), $string,
+            @args{qw( fg bg )},
+            $args{bg_blend}  // TCOD::BKGND_SET,
+            $args{alignment} // TCOD::LEFT,
+        );
+    });
+
+    $ffi->attach( [ printn_rect => 'print_box' ] => [qw(
+        TCOD_console
+        int
+        int
+        int
+        int
+        size_t
+        string
+        TCOD_color*
+        TCOD_color*
+        int
+        int
+    )] => TCOD_error => sub {
+        my ( $xsub, $console, %args ) = @_;
+        my $string = Encode::encode( 'UTF-8', $args{string}, Encode::FB_CROAK );
+
+        $xsub->(
+            $console, @args{qw( x y width height )},
+            length($string), $string,
+            @args{qw( fg bg )},
+            $args{bg_blend}  // TCOD::BKGND_SET,
+            $args{alignment} // TCOD::LEFT,
+        );
+    });
+
+    $ffi->attach( [ draw_frame_rgb => 'draw_frame' ] => [qw(
+        TCOD_console
+        int
+        int
+        int
+        int
+        int[]
+        TCOD_color*
+        TCOD_color*
+        int
+        bool
+    )] => TCOD_error => sub {
+        my ( $xsub, $console, %args ) = @_;
+
+        Carp::croak 'The title parameter is not supported' if exists $args{title};
+
+        my $decoration = $args{decoration} // [ '┌', '─', '┐', '│', ' ', '│', '└', '─', '┘' ];
+
+        $decoration = [ map ord, split //, $decoration ]
+            unless Ref::Util::is_arrayref $decoration;
+
+        Carp::croak 'Frame decoration must have a length of 9. It has a length of ' . @$decoration
+            if @$decoration != 9;
+
+        $xsub->(
+            $console, @args{qw( x y width height )},
+            $decoration,
+            @args{qw( fg bg )},
+            $args{bg_blend}  // TCOD::BKGND_SET,
+            $args{clear}     // 1,
+        );
+    });
+
+    $ffi->attach( [ draw_rect_rgb => 'draw_rect' ] => [qw(
+        TCOD_console
+        int
+        int
+        int
+        int
+        int
+        TCOD_color*
+        TCOD_color*
+        int
+        bool
+    )] => TCOD_error => sub {
+        my ( $xsub, $console, %args ) = @_;
+
+        $xsub->(
+            $console,
+            @args{qw( x y width height ch fg bg )},
+            $args{bg_blend}  // TCOD::BKGND_SET,
+        );
+    });
+
     # Methods
-    $ffi->attach( save_asc      => [qw( TCOD_console string                                           )] => 'bool' );
-    $ffi->attach( save_apf      => [qw( TCOD_console string                                           )] => 'bool' );
-    $ffi->attach( blit          => [qw( TCOD_console int int int int TCOD_console int int float float )] => 'void' );
-    $ffi->attach( set_key_color => [qw( TCOD_console TCOD_color                                       )] => 'void' );
-    $ffi->attach( clear         => [qw( TCOD_console                                                  )] => 'void' );
+
+    my $clear = $bundle->function( clear => [qw( TCOD_console TCOD_color* TCOD_color* int )] );
+    $ffi->attach( clear => [qw( TCOD_console )] => 'void' => sub {
+        my ( $xsub, $console, %args ) = @_;
+
+        return $xsub->($console) unless %args;
+
+        $clear->( $console, $args{fg}, $args{bg}, $args{ch} // TCOD::Event::K_SPACE );
+    });
+
+    $ffi->attach( save_asc                => [qw( TCOD_console string                            )] => 'bool'       );
+    $ffi->attach( save_apf                => [qw( TCOD_console string                            )] => 'bool'       );
+
+    $ffi->attach( set_char                => [qw( TCOD_console int int int                       )] => 'void'       );
+    $ffi->attach( get_char                => [qw( TCOD_console int int                           )] => 'int'        );
+    $ffi->attach( put_char                => [qw( TCOD_console int int int int                   )] => 'void'       );
+    $ffi->attach( put_char_ex             => [qw( TCOD_console int int int TCOD_color TCOD_color )] => 'void'       );
 
     $ffi->attach( get_width               => [qw( TCOD_console                                   )] => 'int'        );
     $ffi->attach( get_height              => [qw( TCOD_console                                   )] => 'int'        );
@@ -526,34 +799,6 @@ package TCOD::Console {
     $ffi->attach( set_default_foreground  => [qw( TCOD_console TCOD_color                        )] => 'void'       );
     $ffi->attach( get_default_foreground  => [qw( TCOD_console                                   )] => 'TCOD_color' );
 
-  # $ffi->attach( get_foreground_color_image => [qw( TCOD_console )] => 'TCOD_image'  );
-  # $ffi->attach( get_background_color_image => [qw( TCOD_console )] => 'TCOD_image'  );
-
-    $ffi->attach( set_char                => [qw( TCOD_console int int int                       )] => 'void'       );
-    $ffi->attach( get_char                => [qw( TCOD_console int int                           )] => 'int'        );
-    $ffi->attach( put_char                => [qw( TCOD_console int int int int                   )] => 'void'       );
-    $ffi->attach( put_char_ex             => [qw( TCOD_console int int int TCOD_color TCOD_color )] => 'void'       );
-
-    $ffi->attach( rect                    => [qw( TCOD_console int int int int bool int          )] => 'void'       );
-    $ffi->attach( hline                   => [qw( TCOD_console int int int          int          )] => 'void'       );
-    $ffi->attach( vline                   => [qw( TCOD_console int int int          int          )] => 'void'       );
-
-    # We support up to two controls per function for now
-    $ffi->attach( print               => [qw( TCOD_console int int                   string )] => [qw( int int )] => 'void' );
-    $ffi->attach( print_ex            => [qw( TCOD_console int int int int           string )] => [qw( int int )] => 'void' );
-    $ffi->attach( print_rect          => [qw( TCOD_console int int int int           string )] => [qw( int int )] => 'void' );
-    $ffi->attach( print_rect_ex       => [qw( TCOD_console int int int int int  int  string )] => [qw( int int )] => 'void' );
-    $ffi->attach( print_frame         => [qw( TCOD_console int int int int bool int  string )] => [qw( int int )] => 'void' );
-    $ffi->attach( get_height_rect     => [qw( TCOD_console int int int int           string )] => [qw( int int )] => 'int'  );
-
-    # UTF-8 variants
-    $ffi->attach( print_utf           => [qw( TCOD_console int int                  wstring )] => [qw( int int )] => 'void' );
-    $ffi->attach( print_ex_utf        => [qw( TCOD_console int int int int          wstring )] => [qw( int int )] => 'void' );
-    $ffi->attach( print_rect_utf      => [qw( TCOD_console int int int int          wstring )] => [qw( int int )] => 'void' );
-    $ffi->attach( print_rect_ex_utf   => [qw( TCOD_console int int int int int int  wstring )] => [qw( int int )] => 'void' );
-  # $ffi->attach( print_frame_utf     => [qw( TCOD_console int int int int bool int wstring )] => [qw( int int )] => 'void' );
-    $ffi->attach( get_height_rect_utf => [qw( TCOD_console int int int int          wstring )] => [qw( int int )] => 'int'  );
-
     # Root console functions
     $ffi->attach( init_root               => [qw( int int string int bool )] => 'void' );
     $ffi->attach( set_custom_font         => [qw( string int int int      )] => 'void' );
@@ -686,11 +931,10 @@ package TCOD::Dijkstra {
         return ( $x, $y );
     });
 
-  # FIXME: Where did this function go?
-  # $ffi->attach( walk => [qw( TCOD_dijkstra int* int* )] => 'bool' => sub {
-  #     shift->( shift, \my $x, \my $y ) or return;
-  #     return ( $x, $y );
-  # });
+    $ffi->attach( [ path_walk => 'walk' ] => [qw( TCOD_dijkstra int* int* )] => 'bool' => sub {
+        $_[0]->( $_[1], \my $x, \my $y, $_[2] ) or return;
+        return ( $x, $y );
+    });
 
     $ffi->mangler( sub { shift } );
     $ffi->attach( [ TCOD_dijkstra_delete => 'DESTROY' ] => ['TCOD_dijkstra'] => 'void' );
@@ -770,6 +1014,181 @@ package TCOD::Noise {
     $ffi->attach( [ TCOD_noise_delete => 'DESTROY' ] => ['TCOD_noise'] => 'void' );
 }
 
+package TCOD::Tileset {
+    $ffi->mangler( sub { shift } );
+
+    FFI::C->struct( TCOD_tileset => [
+        tile_width           => 'int',
+        tile_height          => 'int',
+        tile_length          => 'int',
+        tiles_capacity       => 'int',
+        tiles_count          => 'int',
+        pixels               => 'opaque', # TCOD_ColorRGBA*
+        character_map_length => 'int',
+        character_map        => 'opaque',
+        observer_list        => 'opaque', # TCOD_TilesetObserver
+        virtual_columns      => 'int',
+        ref_count            => 'int',
+    ]);
+
+    sub tile_shape {
+        my $self = shift;
+        ( $self->tile_height, $self->tile_height );
+    }
+
+    $ffi->attach( [ TCOD_load_bdf => 'load_bdf' ] => [qw( string )] => 'TCOD_tileset' => sub {
+        my ( $xsub, $self, $path ) = @_;
+
+        Carp::croak "Cannot load tilesheet from $path: no such file"
+            unless -f $path;
+
+        $xsub->( $self, $path );
+    });
+
+    $ffi->attach( [ TCOD_tileset_load => 'load_tilesheet' ] => [qw( string int int int int[] )] => 'TCOD_tileset' => sub {
+        my ( $xsub, undef, $path, $cols, $rows, $charmap ) = @_;
+
+        Carp::croak "Cannot load tilesheet from $path: no such file"
+            unless -f $path;
+
+        $xsub->( $path, $cols, $rows, scalar @{ $charmap }, $charmap );
+    });
+
+    $ffi->attach( [ TCOD_tileset_get_tile_ => 'get_tile' ] => [qw( TCOD_tileset int )] => 'int' => sub {
+        $_[0]->( $_[1], $_[2], my $color = TCOD::ColorRGBA->new );
+        $color;
+    });
+
+    $ffi->attach( [ TCOD_tileset_set_tile_ => 'set_tile' ] => [qw( TCOD_tileset int TCOD_colorRGBA* )] => 'int' );
+
+    $ffi->attach( [ TCOD_tileset_assign_tile => 'remap' ] => [qw( TCOD_tileset int int int )] => 'int' => sub {
+        my ( $xsub, $self, $x, $y, $codepoint ) = @_;
+        my $i = $x + $y * $self->tileset->virtual_columns;
+
+        Carp::croak "Tile $i is non-existent and can't be assigned"
+            if $i < 0 || $i >= $self->tileset->tiles_count;
+
+        $xsub->( $self, $i, $codepoint );
+    });
+
+     $ffi->attach( [ TCOD_tileset_render_to_surface => 'render' ] => [qw( TCOD_console TCOD_console opaque )] => 'int' => sub {
+        my ( $xsub, $self, $console ) = @_;
+
+        my $width  = $console->width  * $self->tile_width;
+        my $height = $console->height * $self->tile_height;
+        my ($pixels) = FFI::Platypus::Buffer::scalar_to_buffer( pack 'S*', (0) x ( $height * $width ) );
+
+        my $surface = TCOD::SDL2::CreateRGBSurfaceWithFormatFrom(
+            $pixels,
+            $height, $width, 32, 4,
+            TCOD::SDL2::PIXELFORMAT_RGBA32,
+        ) or die TCOD::SDL2::GetError();
+
+        my $pointer = $ffi->cast( opaque => 'opaque*' => $surface );
+
+        $xsub->( $self, $console, undef, $pointer );
+    });
+
+    # $ffi->attach( [ TCOD_tileset_delete => 'DESTROY' ] => ['TCOD_tileset'] => 'void' );
+}
+
+package
+    TCOD::ContextParams {
+    FFI::C->struct( TCOD_context_params => [
+        tcod_version      => 'int',
+        window_x          => 'int',
+        window_y          => 'int',
+        pixel_width       => 'int',
+        pixel_height      => 'int',
+        columns           => 'int',
+        rows              => 'int',
+        renderer_type     => 'int',
+        _tileset          => 'opaque',
+        vsync             => 'int',
+        sdl_window_flags  => 'int',
+        window_title      => 'opaque',
+        argc              => 'int',
+        argv              => 'opaque',
+        cli_output        => 'opaque',
+        cli_userdata      => 'opaque',
+        window_xy_defined => 'bool',
+    ]);
+
+    sub tileset { $ffi->cast( opaque => TCOD_tileset => shift->_tileset ) }
+}
+
+package
+    TCOD::ViewportOptions {
+    FFI::C->struct( TCOD_viewport_options => [
+        tcod_version      => 'int',
+        keep_aspect       => 'bool',
+        integer_scaling   => 'bool',
+        _clear_color      => 'opaque', # TCOD_ColorRGBA
+        align_x           => 'float',
+        align_y           => 'float',
+    ]);
+}
+
+package TCOD::Context {
+    $ffi->mangler( sub { 'TCOD_context_' . shift } );
+
+    $ffi->attach( new  => [qw( TCOD_context_params opaque* )] => 'TCOD_error' => sub {
+        my ( $xsub, undef, %args ) = @_;
+
+        for (
+            [ x        => 'window_x'                        ],
+            [ y        => 'window_y'                        ],
+            [ width    => 'pixel_width'                     ],
+            [ height   => 'pixel_height'                    ],
+            [ renderer => 'renderer_type'                   ],
+            [ title    => 'window_title'  => 'string'       ],
+            [ tileset  => '_tileset'      => 'TCOD_tileset' ],
+        ) {
+            my ( $in, $out, $cast ) = @$_;
+            next unless exists $args{$in};
+
+            $args{$out} = delete $args{$in};
+            $args{$out} = $ffi->cast( $cast => opaque => $args{$out} ) if $cast;
+        }
+
+        my $err = $xsub->( TCOD::ContextParams->new(\%args), \my $ctx );
+        Carp::croak TCOD::get_error() if $err < 0;
+
+        bless \$ctx, 'TCOD::Context';
+    });
+
+    sub new_terminal {
+        my ( $class, $cols, $rows, %args ) = @_;
+
+        $class->new(
+            rows     => $rows,
+            columns  => $cols,
+            vsync    => $args{sync} // 1,
+            %args{qw( renderer tileset sdl_window_flags title )},
+        );
+    }
+
+    $ffi->attach( present => [qw( TCOD_context TCOD_console TCOD_viewport_options )] => 'TCOD_error' => sub {
+        my ( $xsub, $self, $console, %args ) = @_;
+
+        @args{qw( align_x align_y )} = @{ delete $args{align} // [ 0.5, 0.5 ] };
+
+        my $c = delete $args{clear_color} // TCOD::BLACK();
+        $args{_clear_color} = $ffi->cast(
+            TCOD_colorRGBA => opaque => TCOD::ColorRGBA->new( $c->r, $c->g, $c->b, 0xFF ),
+        );
+
+        my $err = $xsub->( $self, $console, TCOD::ViewportOptions->new(\%args) );
+        Carp::croak TCOD::get_error() if $err < 0;
+
+        return;
+    });
+
+
+    $ffi->mangler( sub { shift } );
+    $ffi->attach( [ TCOD_context_delete => 'DESTROY' ] => ['TCOD_context'] => 'void' );
+}
+
 for (
     # color values
     [ BLACK                  => [   0,   0,   0 ] ],
